@@ -4,12 +4,13 @@
 use crate::arch::set_thread_pointer;
 use crate::arch::{clone, get_thread_pointer, munmap_and_exit_thread, TLS_OFFSET};
 use memoffset::offset_of;
-use rsix::io;
+use rustix::io;
 #[cfg(target_vendor = "mustang")]
-use rsix::process::{getrlimit, linux_execfn, Resource};
-use rsix::process::{page_size, Pid};
-use rsix::runtime::{set_tid_address, StartupTlsInfo};
-use rsix::thread::gettid;
+use rustix::process::{getrlimit, linux_execfn, Resource};
+use rustix::process::{page_size, Pid};
+use rustix::runtime::{set_tid_address, StartupTlsInfo};
+use rustix::thread::gettid;
+use std::any::Any;
 use std::cmp::max;
 use std::ffi::c_void;
 use std::mem::{align_of, size_of};
@@ -27,16 +28,10 @@ use std::sync::atomic::{AtomicU32, AtomicU8};
 ///
 /// After calling `func`, this terminates the thread.
 #[cfg(feature = "threads")]
-pub(super) unsafe extern "C" fn entry(
-    arg: *mut c_void,
-    func: unsafe extern "C" fn(*mut c_void) -> *mut c_void,
-) -> ! {
-    log::trace!(
-        "Thread[{:?}] launched; calling `{:?}({:?})`",
-        current_thread_id(),
-        func,
-        arg
-    );
+pub(super) unsafe extern "C" fn entry(fn_: *mut Box<dyn FnOnce() -> Option<Box<dyn Any>>>) -> ! {
+    let fn_ = Box::from_raw(fn_);
+
+    log::trace!("Thread[{:?}] launched", current_thread_id());
 
     // Do some basic precondition checks, to ensure that our assembly code did
     // what we expect it to do. These are debug-only for now, to keep the
@@ -75,7 +70,7 @@ pub(super) unsafe extern "C" fn entry(
 
     // Call the user thread function. In `std`, this is `thread_start`. Ignore
     // the return value for now, as `std` doesn't need it.
-    let _result = func(arg);
+    let _result = fn_();
 
     exit_thread()
 }
@@ -127,7 +122,7 @@ pub struct Thread {
     stack_size: usize,
     guard_size: usize,
     map_size: usize,
-    dtors: Vec<(unsafe extern "C" fn(*mut c_void), *mut c_void)>,
+    dtors: Vec<Box<dyn FnOnce()>>,
 }
 
 // Values for `Thread::detached`.
@@ -172,7 +167,7 @@ pub fn current_thread() -> *mut Thread {
 
 /// Return the current thread id.
 ///
-/// This is the same as `rsix::thread::gettid()`, but loads the value from a
+/// This is the same as `rustix::thread::gettid()`, but loads the value from a
 /// field in the runtime rather than making a system call.
 #[inline]
 pub fn current_thread_id() -> Pid {
@@ -180,6 +175,12 @@ pub fn current_thread_id() -> Pid {
     debug_assert_eq!(tid, gettid(), "`current_thread_id` disagrees with `gettid`");
     tid
 }
+/// Change the current thread id.
+#[inline]
+pub(crate) fn set_current_thread_id(tid: Pid) {
+    unsafe { (*current_thread()).thread_id.store(tid.as_raw(), SeqCst) };
+    debug_assert_eq!(tid, gettid(), "`current_thread_id` disagrees with `gettid`");
+}
 
 /// Return the TLS entry for the current thread.
 #[inline]
@@ -225,23 +226,30 @@ pub unsafe fn thread_stack(thread: *mut Thread) -> (*mut c_void, usize, usize) {
 ///
 /// This arranges for `func` to be called, and passed `obj`, when the thread
 /// exits.
-pub unsafe fn at_thread_exit(func: unsafe extern "C" fn(*mut c_void), obj: *mut c_void) {
-    let current = current_thread();
-    (*current).dtors.push((func, obj));
+pub fn at_thread_exit(func: Box<dyn FnOnce()>) {
+    // Safety: `current_thread()` points to thread-local data which is valid
+    // as long as the thread is alive.
+    unsafe {
+        (*current_thread()).dtors.push(func);
+    }
 }
 
 /// Call the destructors registered with `at_thread_exit`.
-unsafe fn call_thread_dtors(current: *mut Thread) {
+fn call_thread_dtors(current: *mut Thread) {
     // Run the `dtors`, in reverse order of registration. Note that destructors
     // may register new destructors.
-    while let Some((func, obj)) = (*current).dtors.pop() {
-        log::trace!(
-            "Thread[{:?}] calling `at_thread_exit`-registered function `{:?}({:?})`",
-            (*current).thread_id,
-            func,
-            obj
-        );
-        func(obj);
+    //
+    // Safety: `current` points to thread-local data which is valid as long
+    // as the thread is alive.
+    while let Some(func) = unsafe { (*current).dtors.pop() } {
+        if log::log_enabled!(log::Level::Trace) {
+            log::trace!(
+                "Thread[{:?}] calling `at_thread_exit`-registered function",
+                unsafe { (*current).thread_id.load(SeqCst) },
+            );
+        }
+
+        func();
     }
 }
 
@@ -295,7 +303,7 @@ unsafe fn exit_thread() -> ! {
     }
 
     // Terminate the thread.
-    rsix::runtime::exit_thread(0)
+    rustix::runtime::exit_thread(0)
 }
 
 /// Initialize the main thread.
@@ -306,7 +314,7 @@ unsafe fn exit_thread() -> ! {
 #[cfg(target_vendor = "mustang")]
 pub(super) unsafe fn initialize_main_thread(mem: *mut c_void) {
     // Read the TLS information from the ELF header.
-    STARTUP_TLS_INFO = rsix::runtime::startup_tls_info();
+    STARTUP_TLS_INFO = rustix::runtime::startup_tls_info();
 
     // Determine the top of the stack. Linux puts the `AT_EXECFN` string at
     // the top, so find the end of that, and then round up to the page size.
@@ -417,8 +425,7 @@ pub(super) unsafe fn initialize_main_thread(mem: *mut c_void) {
 /// calls `fn_`, which does, and it's marked `unsafe`.
 #[allow(clippy::not_unsafe_ptr_arg_deref)]
 pub fn create_thread(
-    fn_: unsafe extern "C" fn(*mut c_void) -> *mut c_void,
-    arg: *mut c_void,
+    fn_: Box<dyn FnOnce() -> Option<Box<dyn Any>>>,
     stack_size: usize,
     guard_size: usize,
 ) -> io::Result<*mut Thread> {
@@ -567,8 +574,7 @@ pub fn create_thread(
             thread_id_ptr,
             thread_id_ptr,
             newtls.cast::<u8>().cast(),
-            arg,
-            fn_,
+            Box::into_raw(Box::new(fn_)),
         );
         #[cfg(any(
             target_arch = "x86",
@@ -582,8 +588,7 @@ pub fn create_thread(
             thread_id_ptr,
             newtls.cast::<u8>().cast(),
             thread_id_ptr,
-            arg,
-            fn_,
+            Box::into_raw(Box::new(fn_)),
         );
         if clone_res >= 0 {
             Ok(&mut (*metadata).thread)
@@ -641,7 +646,7 @@ pub unsafe fn join_thread(thread: *mut Thread) {
 }
 
 unsafe fn wait_for_thread_exit(thread: *mut Thread) {
-    use rsix::thread::{futex, FutexFlags, FutexOperation};
+    use rustix::thread::{futex, FutexFlags, FutexOperation};
 
     // Check whether the thread has exited already; we set the
     // `CloneFlags::CHILD_CLEARTID` flag on the clone syscall, so we can test
@@ -670,7 +675,7 @@ unsafe fn wait_for_thread_exit(thread: *mut Thread) {
 }
 
 unsafe fn free_thread_memory(thread: *mut Thread) {
-    use rsix::io::munmap;
+    use rustix::io::munmap;
 
     // The thread was detached. Prepare to free the memory. First read out
     // all the fields that we'll need before freeing it.
@@ -725,7 +730,7 @@ unsafe extern "C" fn __aeabi_read_tp() -> *mut c_void {
     get_thread_pointer()
 }
 
-// We define `clone` and `CloneFlags` here in `origin` instead of `rsix`
+// We define `clone` and `CloneFlags` here in `origin` instead of `rustix`
 // because `clone` needs custom assembly code that knows about what we're
 // using it for.
 bitflags::bitflags! {
